# HiQ version 1.1.2
#
# Copyright (c) 2022, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
#

import time

import hiq
from hiq import HiQLatency, HiQMemory
from hiq.constants import *
from hiq.hiq_utils import func_args_handler
from hiq.memory import get_memory_mb

from typing import Union, Callable, List
from hiq.hiq_utils import (
    func_args_handler,
    get_tau_id,
)


class FastAPIReqIdGenerator(object):
    def __call__(self):
        from asgi_correlation_id.context import correlation_id

        return correlation_id.get()


def run_fastapi(app, driver=None, header_name='X-Request-ID', host='0.0.0.0', port=8080, worker=1):
    from asgi_correlation_id import CorrelationIdMiddleware
    from uuid import uuid4
    # from asgi_correlation_id.middleware import is_valid_uuid4

    app.add_middleware(CorrelationIdMiddleware,
                       header_name=header_name,
                       generator=lambda: uuid4().hex,
                       # validator=is_valid_uuid4
                       )

    @app.get("/hiq")
    async def hiq_report():
        from fastapi.responses import HTMLResponse
        if not driver:
            return ""
        r = driver.get_k0s_summary()
        tmp=[]
        for k, span, start, end in r:
            s = f'<tr><td><a href=latency/{k}>🟢 {k}</a><td align=right>{span}<td align=right>{hiq.ts_to_dt(start)}<td align=right>{hiq.ts_to_dt(end)}'
            tmp.append(s)
        resp = '<table width=100%><tr><td width=50% align=left>Req ID<td align=right>Latency<td align=right>Start<td align=right>End' + '\n'.join(tmp) + "</table>"
        html_resp = f"""
                <!DOCTYPE html>
                <html>
                  <head>
                      <link href="https://unpkg.com/tailwindcss@^2/dist/tailwind.min.css" rel="stylesheet">
                  </head>
                  <body class="bg-white">
                    <section style='font-family: monospace, Times, serif;'>
                        <div class="container mx-auto">
                            <div class="flex flex-col flex-wrap pb-6 mb-2 text-black ">
                                <h1 class="mb-2 text-3xl font-medium text-black">
                                    Request Latency Report
                                </h1>
                                <p class="text-xl leading-relaxed"> HiQ </p> </div> <div class="flex flex-wrap 
                                items-end justify-start w-full transition duration-500 ease-in-out transform bg-white 
                                border-2 border-gray-600 rounded-lg hover:border-white "> <div class="w-full"> <div 
                                class="relative flex flex-col h-full p-8 text-sm"> <pre>{resp}</pre>
                                </div>
                            </div>
                        </div>
                    </section>
                  </body>
                </html>
                """
        return HTMLResponse(content=html_resp, status_code=200)

    @app.get("/latency/{req_id}")
    async def latency(req_id: str):
        if not driver:
            return ""
        from fastapi.responses import HTMLResponse

        t = driver.get_metrics_by_k0(req_id)
        resp = t.get_graph(FORMAT_DATETIME) if t else f"Error: The reqeust id({req_id}) is invalid!"
        html_resp = f"""
        <!DOCTYPE html>
        <html>
          <head>
              <link href="https://unpkg.com/tailwindcss@^2/dist/tailwind.min.css" rel="stylesheet">
          </head>
          <body class="bg-white">
            <section style='font-family: monospace, Times, serif;'>
                <div class="container mx-auto">
                    <div class="flex flex-col flex-wrap pb-6 mb-2 text-black ">
                        <h1 class="mb-2 text-3xl font-medium text-black">
                            Request Latency Report
                        </h1>
                        <p class="text-xl leading-relaxed">Request Id: {req_id}</p>
                    </div>
                    <div class="flex flex-wrap 
                        items-end justify-start w-full transition duration-500 ease-in-out transform bg-white 
                        border-2 border-gray-600 rounded-lg hover:border-white "> <div class="w-full"> <div 
                        class="relative flex flex-col h-full p-8 text-sm"> <pre>{resp}</pre> </div> </div> <div 
                        class="w-full xl:w-1/4 md:w-1/4 lg:ml-auto"> <div class="relative flex flex-col h-full p-8"> 
                        <button class="w-full px-4 py-2 mx-auto mt-3"> <a href="../hiq">Back to HiQ Report Home 🏠</a> </button> </div> </div> 
                        </div> </section> </body> </html>"""
        return HTMLResponse(content=html_resp, status_code=200)

    import uvicorn
    uvicorn.run(app, host=host, port=port, workers=worker)


def get_corr_id():
    from asgi_correlation_id.context import correlation_id
    return correlation_id.get()


class HiQFastAPILatencyMixin:
    def __init__(sf,
                 hiq_table_or_path: Union[str, list] = [],
                 metric_funcs: List[Callable] = [time.time],
                 hiq_id_func: Callable = FastAPIReqIdGenerator(),
                 func_args_handler: Callable = func_args_handler,
                 target_path=None,
                 max_hiq_size=30,
                 verbose=False,
                 fast_fail=True,
                 tpl=None,
                 extra_hiq_table=[],
                 attach_timestamp=False,
                 extra_metrics=set(),
                 lmk_path=None,
                 lmk_handler=None,
                 lmk_logger=None,
                 ):
        super().__init__(hiq_id_func=hiq_id_func,
                         hiq_table_or_path=hiq_table_or_path,
                         metric_funcs=metric_funcs,
                         func_args_handler=func_args_handler,
                         target_path=target_path,
                         max_hiq_size=max_hiq_size,
                         verbose=verbose,
                         fast_fail=fast_fail,
                         tpl=tpl,
                         extra_hiq_table=extra_hiq_table,
                         attach_timestamp=attach_timestamp,
                         extra_metrics=extra_metrics,
                         lmk_path=lmk_path,
                         lmk_handler=lmk_handler,
                         lmk_logger=lmk_logger,
                         )


class HiQFastAPILatency(HiQLatency):
    def __init__(
            sf,
            hiq_table_or_path: Union[str, list] = [],
            metric_funcs: List[Callable] = [time.time],
            func_args_handler: Callable = func_args_handler,
            target_path=None,
            max_hiq_size=30,
            verbose=False,
            fast_fail=True,
            tpl=None,
            extra_hiq_table=[],
            attach_timestamp=False,
            extra_metrics=set(),
            lmk_path=None,
            lmk_handler=None,
            lmk_logger=None,
    ):
        HiQLatency.__init__(
            sf,
            hiq_table_or_path=hiq_table_or_path,
            metric_funcs=metric_funcs,
            hiq_id_func=FastAPIReqIdGenerator(),
            func_args_handler=func_args_handler,
            target_path=target_path,
            max_hiq_size=max_hiq_size,
            verbose=verbose,
            fast_fail=fast_fail,
            tpl=tpl,
            extra_hiq_table=extra_hiq_table,
            attach_timestamp=attach_timestamp,
            extra_metrics=extra_metrics,
            lmk_path=lmk_path,
            lmk_handler=lmk_handler,
            lmk_logger=lmk_logger,
        )

    def custom(sf):
        pass

    def custom_disable(sf):
        pass


class HiQFastAPIMemory(HiQMemory):
    def __init__(
            sf,
            hiq_table_or_path: Union[str, list] = [],
            metric_funcs: List[Callable] = [time.time, get_memory_mb],
            func_args_handler: Callable = func_args_handler,
            target_path=None,
            max_hiq_size=30,
            verbose=False,
            fast_fail=True,
            tpl=None,
            extra_hiq_table=[],
            attach_timestamp=False,
            extra_metrics=set(),
            lmk_path=None,
            lmk_handler=None,
            lmk_logger=None,
    ):
        HiQLatency.__init__(
            sf,
            hiq_table_or_path=hiq_table_or_path,
            metric_funcs=metric_funcs,
            hiq_id_func=FastAPIReqIdGenerator(),
            func_args_handler=func_args_handler,
            target_path=target_path,
            max_hiq_size=max_hiq_size,
            verbose=verbose,
            fast_fail=fast_fail,
            tpl=tpl,
            extra_hiq_table=extra_hiq_table,
            attach_timestamp=attach_timestamp,
            extra_metrics=extra_metrics,
            lmk_path=lmk_path,
            lmk_handler=lmk_handler,
            lmk_logger=lmk_logger,
        )

    def custom(sf):
        pass

    def custom_disable(sf):
        pass



import hiq
import os
from typing import *
import time
from hiq.memory import get_memory_mb
from hiq.hiq_utils import (
    func_args_handler,
    get_tau_id,
)
from hiq.utils import get_env_int

TAU_ORT = "onnxruntime.capi.onnxruntime_inference_collection"

TAU_TBL_ORT = [
    (TAU_ORT, "Session", "run", "ort_run"),
    (TAU_ORT, "Session", "__init__", "sess_init"),
    (TAU_ORT, "Session", "run_with_ort_values", "run_with_ort_values"),
    (TAU_ORT, "Session", "run_with_iobinding", "run_with_iobinding"),
    (TAU_ORT, "InferenceSession", "_create_inference_session", "create_inference_session"),
    (TAU_ORT, "InferenceSession", "_reset_session", "reset_session"),
]

def get_ort_session(sess_options):
    if "HIQ_ORT_INTRA_OPS_THREAD" in os.environ:
        import onnxruntime as ort
        if sess_options is None:
            sess_options = ort.SessionOptions()
        sess_options.intra_op_num_threads = get_env_int("HIQ_ORT_INTRA_OPS_THREAD")
    return sess_options

class OrtHiQLatency(hiq.HiQSimple):
    
    def __init__(
        sf,
        hiq_table_or_path: Union[str, list] = [],
        metric_funcs: List[Callable] = [time.time],
        hiq_id_func: Callable = get_tau_id,
        func_args_handler: Callable = func_args_handler,
        target_path=None,
        max_hiq_size=30,
        verbose=False,
        fast_fail=True,
        tpl=None,
        extra_hiq_table=[],
        attach_timestamp=False,
        extra_metrics=set(),
        lmk_path=None,
        lmk_handler=None,
        lmk_logger=None,
    ):
        extra_hiq_table += TAU_TBL_ORT
        hiq.HiQSimple.__init__(
            sf,
            hiq_table_or_path=hiq_table_or_path,
            metric_funcs=metric_funcs,
            hiq_id_func=hiq_id_func,
            func_args_handler=func_args_handler,
            target_path=target_path,
            max_hiq_size=max_hiq_size,
            verbose=verbose,
            fast_fail=fast_fail,
            tpl=tpl,
            extra_hiq_table=extra_hiq_table,
            attach_timestamp=attach_timestamp,
            extra_metrics=extra_metrics,
            lmk_path=lmk_path,
            lmk_handler=lmk_handler,
            lmk_logger=lmk_logger,
        )
        
    
    def custom(s):
        s.o_ort_session = hiq.mod(TAU_ORT).InferenceSession.__init__

        @s.inserter
        def ort_session(self,
                path_or_bytes,
                sess_options=None,
                providers=None,
                provider_options=None):
            return s.o_ort_session(
                    self, path_or_bytes, get_ort_session(sess_options), providers, provider_options
                )
        hiq.mod(TAU_ORT).InferenceSession.__init__ = ort_session

    def custom_disable(s):
        hiq.mod(TAU_ORT).InferenceSession.__init__ = s.o_ort_session


class OrtHiQMemory(hiq.HiQMemory):
    
    def __init__(
        sf,
        hiq_table_or_path: Union[str, list] = [],
        metric_funcs: List[Callable] = [time.time, get_memory_mb],
        hiq_id_func: Callable = get_tau_id,
        func_args_handler: Callable = func_args_handler,
        target_path=None,
        max_hiq_size=30,
        verbose=False,
        fast_fail=True,
        tpl=None,
        extra_hiq_table=[],
        attach_timestamp=False,
        extra_metrics=set(),
        lmk_path=None,
        lmk_handler=None,
        lmk_logger=None,
    ):
        extra_hiq_table += TAU_TBL_ORT
        hiq.HiQSimple.__init__(
            sf,
            hiq_table_or_path=hiq_table_or_path,
            metric_funcs=metric_funcs,
            hiq_id_func=hiq_id_func,
            func_args_handler=func_args_handler,
            target_path=target_path,
            max_hiq_size=max_hiq_size,
            verbose=verbose,
            fast_fail=fast_fail,
            tpl=tpl,
            extra_hiq_table=extra_hiq_table,
            attach_timestamp=attach_timestamp,
            extra_metrics=extra_metrics,
            lmk_path=lmk_path,
            lmk_handler=lmk_handler,
            lmk_logger=lmk_logger,
        )
    
    def custom(s):
        s.o_ort_session = hiq.mod(TAU_ORT).InferenceSession.__init__

        @s.inserter
        def ort_session(self,
                path_or_bytes,
                sess_options=None,
                providers=None,
                provider_options=None):
            return s.o_ort_session(
                    self, path_or_bytes, get_ort_session(sess_options), providers, provider_options
                )
        hiq.mod(TAU_ORT).InferenceSession.__init__ = ort_session

    def custom_disable(s):
        hiq.mod(TAU_ORT).InferenceSession.__init__ = s.o_ort_session



# HiQ version 1.1.2
#
# Copyright (c) 2022, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/ 
#

from enum import Enum
import os
from typing import Optional, Sequence
import itree
from hiq.hiq_utils import TRACING_TYPE_OTM
from hiq.utils import memoize_first


class OtmExporterType(Enum):
    JAEGER_THRIFT = 1
    JAEGER_PROTOBUF = 2
    ZIPKIN_JSON = 3
    ZIPKIN_PROTOBUF = 4


def setup_jaeger(
    service_name="hiq_service",
    agent_host_name="localhost",
    agent_port=6831,
    tracer_name="otm_hiq",
    collector_endpoint="localhost:14250",
    exporter_type: OtmExporterType = OtmExporterType.JAEGER_THRIFT,
):
    """Set up jaeger client

    Args:
        service_name (str, optional): the name of the service. Defaults to "hiq_service".
        agent_host_name (str, optional): jaeger agent host name. Defaults to "localhost".
        agent_port (int, optional): jaeger agent port. Defaults to 6831.
        tracer_name (str, optional): tracer name. Defaults to "otm_hiq".
    """

    from opentelemetry import trace
    from opentelemetry.sdk.resources import SERVICE_NAME, Resource
    from opentelemetry.sdk.trace import TracerProvider
    from opentelemetry.sdk.trace.export import BatchSpanProcessor

    trace.set_tracer_provider(
        TracerProvider(resource=Resource.create({SERVICE_NAME: service_name}))
    )
    if exporter_type == OtmExporterType.JAEGER_THRIFT:
        jaeger_exporter = itree.mod(
            "opentelemetry.exporter.jaeger.thrift"
        ).JaegerExporter(
            agent_host_name=agent_host_name,
            agent_port=agent_port,
        )
    elif exporter_type == OtmExporterType.JAEGER_PROTOBUF:
        jaeger_exporter = itree.mod(
            "opentelemetry.exporter.jaeger.proto.grpc"
        ).JaegerExporter(collector_endpoint=collector_endpoint, insecure=True)
    else:
        raise ValueError(f"wrong exporter type: {exporter_type}")

    trace.get_tracer_provider().add_span_processor(BatchSpanProcessor(jaeger_exporter))

    os.environ["OTM_TRACER_NAME"] = tracer_name


def setup_zipkin(
    service_name="hiq_service",
    tracer_name="otm_hiq",
    endpoint: Optional[str] = "http://localhost:9411/api/v2/spans",
    local_node_ipv4=None,
    local_node_port: Optional[int] = None,
    max_tag_value_length: Optional[int] = 256,
    timeout: Optional[int] = 5,
    exporter_type: OtmExporterType = OtmExporterType.ZIPKIN_JSON,
):
    from opentelemetry import trace
    from opentelemetry.sdk.resources import SERVICE_NAME, Resource
    from opentelemetry.exporter.zipkin.encoder import Protocol
    from opentelemetry.sdk.trace import TracerProvider
    from opentelemetry.sdk.trace.export import BatchSpanProcessor

    trace.set_tracer_provider(
        TracerProvider(resource=Resource.create({SERVICE_NAME: service_name}))
    )

    if exporter_type == OtmExporterType.ZIPKIN_JSON:
        zipkin_exporter = itree.mod(
            "opentelemetry.exporter.zipkin.json"
        ).ZipkinExporter(
            version=Protocol.V2,
            endpoint=endpoint,
            local_node_ipv4=local_node_ipv4,
            local_node_port=local_node_port,
            max_tag_value_length=max_tag_value_length,
            timeout=timeout,  # (in seconds)
        )
    elif exporter_type == OtmExporterType.ZIPKIN_PROTOBUF:
        zipkin_exporter = itree.mod(
            "opentelemetry.exporter.zipkin.proto.http"
        ).ZipkinExporter(
            endpoint=endpoint,
            local_node_ipv4=local_node_ipv4,
            local_node_port=local_node_port,
            max_tag_value_length=max_tag_value_length,
            timeout=timeout,  # (in seconds)
        )
    else:
        raise ValueError(f"wrong exporter type: {exporter_type}")

    # Create a BatchSpanProcessor and add the exporter to it
    span_processor = BatchSpanProcessor(zipkin_exporter)

    # add to the tracer
    trace.get_tracer_provider().add_span_processor(span_processor)
    os.environ["OTM_TRACER_NAME"] = tracer_name


class HiQOpenTelemetryContext(object):
    """OpenTelemetry Context Manager with HiQ

    Example:

    .. highlight:: python
    .. code-block:: python

        import os

        import hiq
        from hiq.distributed import HiQOpenTelemetryContext, OtmExporterType

        here = os.path.dirname(os.path.realpath(__file__))


        def run_main():
            with HiQOpenTelemetryContext(
                exporter_type=OtmExporterType.JAEGER_PROTOBUF,
                collector_endpoint="localhost:14250"
            ):
                driver = hiq.HiQLatency(f"{here}/hiq.conf")
                hiq.mod("main").main()
                driver.show()


        if __name__ == "__main__":
            run_main()
    """

    from opentelemetry import trace

    o_get_tracer = trace.get_tracer

    @staticmethod
    @memoize_first
    def get_tracer_by_name(name):
        return HiQOpenTelemetryContext.o_get_tracer(name)

    def __init__(
        self,
        exporter_type: OtmExporterType = OtmExporterType.JAEGER_THRIFT,
        *args,
        **kwargs,
    ):
        self.exporter_type = exporter_type
        if self.exporter_type in [
            OtmExporterType.JAEGER_THRIFT,
            OtmExporterType.JAEGER_PROTOBUF,
        ]:
            kwargs["exporter_type"] = self.exporter_type
            setup_jaeger(*args, **kwargs)
        elif self.exporter_type in [
            OtmExporterType.ZIPKIN_JSON,
            OtmExporterType.ZIPKIN_PROTOBUF,
        ]:
            kwargs["exporter_type"] = self.exporter_type
            setup_zipkin(*args, **kwargs)
        else:
            raise ValueError("Unknown exporter type")

    def __enter__(self):
        """To enable OpenTelemetry support, you need to set environment variable TRACE_TYPE equal to opentelemetry."""
        os.environ["TRACE_TYPE"] = TRACING_TYPE_OTM
        self.old_get_tracer = HiQOpenTelemetryContext.trace.get_tracer
        HiQOpenTelemetryContext.trace.get_tracer = (
            HiQOpenTelemetryContext.get_tracer_by_name
        )
        tracer_name = os.environ.get("OTM_TRACER_NAME", "otm_hiq")
        HiQOpenTelemetryContext.trace.get_tracer(tracer_name)

    def __exit__(self, exc_type, exc_value, exc_tb):
        HiQOpenTelemetryContext.trace.get_tracer = self.old_get_tracer
        del os.environ["TRACE_TYPE"]



# HiQ version 1.1.2
#
# Copyright (c) 2022, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/ 
#

from enum import Enum

EXTRA_START_TIME_KEY = "start"
EXTRA_END_TIME_KEY = "end"

TREE_ROOT_NAME = "root"
FORMAT_TIMESTAMP = "_timestamp_"
FORMAT_DATETIME = "_datetime_"

HIQ_FUNC_PREFIX = "__"

HIQ_FUNC_DIO_RD = "dio_r"
HIQ_FUNC_DIO_WT = "dio_w"

HIQ_FUNC_SIO_RD = "sio_r"
HIQ_FUNC_SIO_WT = "sio_w"

HIQ_FUNC_NIO_GET = "nio_get"
HIQ_FUNC_NIO_POS = "nio_pos"

TAU_TABLE_DIO_RD = ["_io", "TextIOWrapper", "read", HIQ_FUNC_DIO_RD]
TAU_TABLE_DIO_WT = ["_io", "TextIOWrapper", "write", HIQ_FUNC_DIO_WT]

HIQ_TABLE_SIO_RD = ["os", "", "read", HIQ_FUNC_SIO_RD]
HIQ_TABLE_SIO_WT = ["os", "", "write", HIQ_FUNC_SIO_WT]

TAU_TABLE_NIO_GET = ["requests", "", "get", HIQ_FUNC_NIO_GET]
TAU_TABLE_NIO_POS = ["requests", "", "post", HIQ_FUNC_NIO_POS]

KEY_LATENCY = "time"
KEY_MEMORY = "get_memory_mb"

KEY_EXC_SUM = "exception_summary"
KEY_EXC_TRA = "exception_trace"

TAU_TPL_LOC = "data/tau.tpl"
TAU_PKL_LOC = "data/tau.pk"

TS_LEN = len(str(1631339134.691))  # 14

EXTERNAL_C_MODULE = "ctypes"

TRACING_TYPE_HIQ = "oci-hiq"
TRACING_TYPE_OCI = "oci-apm"
TRACING_TYPE_OTM = "opentelemetry"
TRACING_TYPE_PROM = "prometheus"

# 1 Page, 32K slots for our hash table
HIQ_SHARED_MEMORY = 1024 * 4
GLOBAL_HIQ_TTL_S = 60

HIQ_C_SET = {"_io.TextIOWrapper.read", "_io.TextIOWrapper.write"}


class ExtraMetrics(Enum):
    ARGS = 1
    FILE = 2
    FUNC = 3


class OverHeadFormat(Enum):
    ABS = 1
    PCT = 2



# HiQ version 1.1.2
#
# Copyright (c) 2022, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/ 
#

from hiq.http_metric_client import HttpMetricsClient


class OciT2Client(HttpMetricsClient):
    """
    OciT2Client is a class for transmitting metrics to Oracle T2

    Examples:

    .. highlight:: python
    .. code-block:: python

        from hiq.vendor_oci_t2 import OciT2Client as Client
        from time import monotonic

        client = Client(url=...)
        METRIC_NAME = "hiq.predict"
        metric = Client.metric_calc_delta(f"{METRIC_NAME}.success", start_time=monotonic())
        metrics_queue.put_nowait(metric)
        #...
        retry_count=1
        client.gauge_metric("operation_retry_count", retry_count)
        client.submit_metrics_queue()

    """



from setuptools import setup
import os

here = os.path.dirname(os.path.realpath(__file__))

VERSION = (
    "1.1.2"
    if "PKG_VERSION" not in os.environ or not os.environ["PKG_VERSION"]
    else os.environ["PKG_VERSION"]
)
DESCRIPTION = "HiQ - A Modern Observability System"

packages = [
    "hiq",
]


def read_file(filename: str):
    try:
        lines = []
        with open(filename) as file:
            lines = file.readlines()
            lines = [line.rstrip() for line in lines if not line.startswith('#')]
        return lines
    except:
        pass

LONG_DESCRIPTION = """
HiQ - A Modern Observability System
*************************************************

|Documentation Status| |CodeCov| |Github release| |lic|

HiQ is a ``declarative``, ``non-intrusive``, ``dynamic`` and
``transparent`` tracking system for both **monolithic** application and
**distributed** system. It brings the runtime information tracking and
optimization to a new level without compromising with speed and system
performance, or hiding any tracking overhead information. HiQ applies
for both I/O bound and CPU bound applications.

To explain the four features, **declarative** means you can declare the
things you want to track in a text file, which could be a json, yaml or
even csv,and no need to change program code. **Non-intrusive** means HiQ
doesn’t requires to modify original python code. **Dynamic** means HiQ
supports tracing metrics featuring at run time, which can be used for
adaptive tracing. **Transparent** means HiQ provides the tracing
overhead and doesn’t hide it no matter it is huge or tiny.

In addition to latency tracking, HiQ provides memory, disk I/O and
Network I/O tracking out of the box. The output can be saved in form of
normal line by line log file, or HiQ tree, or span graph.

Installation
------------

.. code:: bash

   pip install hiq-python

Get Started
-----------

Let start with a simplest example by running HiQ against a simple
monolithic python code `📄
main.py <https://github.com/oracle-samples/hiq/blob/main/hiq/examples/quick_start/main.py>`__:

.. code:: python

   # this is the main.py python source code
   import time

   def func1():
       time.sleep(1.5)
       print("func1")
       func2()

   def func2():
       time.sleep(2.5)
       print("func2")

   def main():
       func1()

   if __name__ == "__main__":
       main()

In this code, there is a simple chain of function calls: ``main()`` ->
``func1`` -> ``func2``.

Now we want to trace the functions without modifying its code. Let’s run
the following:

.. code:: python

   git clone https://github.com/oracle-samples/hiq.git
   cd hiq/examples/quick_start
   python main_driver.py

If everything is fine, you should be able to see the output like this:

.. figure:: https://raw.githubusercontent.com/oracle-samples/hiq/main/hiq/docs/source/img/main_driver.jpg

   An HiQ Simplest Example

From the screenshot we can see the timestamp and the latency of each
function:

=============== ====== ====== ====== ================
\               main   func1  func2  tracing overhead
=============== ====== ====== ====== ================
latency(second) 4.0045 4.0044 2.5026 0.0000163
=============== ====== ====== ====== ================

HiQ just traced the ``main.py`` file running without touching one line
of its code.

Documentation
-------------

**HTML**: `🔗 HiQ Online Document <https://hiq.readthedocs.io>`__, **PDF**: `🔗 HiQ PDF Document <https://github.com/oracle-samples/hiq/raw/main/hiq/docs/hiq.pdf>`__.

Examples
--------

Please check `🔗 examples <https://github.com/oracle-samples/hiq/tree/main/hiq/examples>`__ for usage examples.

Contributing
------------

HiQ welcomes contributions from the community. Before submitting a pull
request, please `review our 🔗 contribution guide <https://github.com/oracle-samples/hiq/blob/main/CONTRIBUTING.md>`__.

Security
--------

Please consult the `🔗 security guide <https://github.com/oracle-samples/hiq/blob/main/SECURITY.md>`__ for our
responsible security vulnerability disclosure process.

License
-------

Copyright (c) 2022 Oracle and/or its affiliates. Released under the
Universal Permissive License v1.0 as shown at
https://oss.oracle.com/licenses/upl/.

.. |image0| image:: https://raw.githubusercontent.com/oracle-samples/hiq/main/hiq/docs/source/_static/hiq.png
.. |Documentation Status| image:: https://readthedocs.org/projects/hiq/badge/?version=latest
   :target: https://hiq.readthedocs.io/en/latest/
.. |CodeCov| image:: https://codecov.io/gh/uber/athenadriver/branch/master/graph/badge.svg
   :target: https://github.com/oracle-samples/hiq
.. |Github release| image:: https://img.shields.io/badge/release-v1.0.3-red
   :target: https://github.com/oracle-samples/hiq
.. |lic| image:: https://img.shields.io/badge/License-UPL--1.0-red
   :target: https://github.com/oracle-samples/hiq/blob/main/LICENSE.txt
"""


setup(
    name="hiq-python",
    version=VERSION,
    author="Henry Fuheng Wu; Kathan Patel",
    author_email="<fuheng.wu@oracle.com>",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    install_requires=read_file(f"{here}/requirements.txt"),
    keywords=[
        "hiq",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    packages=packages,
    package_dir={"": "src"},
    package_data={"hiq": ["data/*.pk", "framework/*"]},
    url="https://github.com/oracle-samples/hiq",
)



# HiQ version 1.1.2
#
# Copyright (c) 2022, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/ 
#

import os
from multiprocessing import Process, Queue, Lock
import itree
from hiq.tree import Tree
from hiq.utils import _check_overhead, get_env_bool, ensure_folder, get_home
import time


def log_jack_rotated(
    log_file=f"{get_home()}/.hiq/log_jack.log",
    max_bytes=500 * 1024 * 1024,
    backup_count=20,
):
    if get_env_bool("NO_JACK_LOG"):
        return None
    import logging
    from logging.handlers import RotatingFileHandler

    ensure_folder(log_file)

    my_handler = RotatingFileHandler(
        log_file,
        mode="a",
        maxBytes=max_bytes,
        backupCount=backup_count,
        encoding=None,
        delay=0,
    )
    my_handler.setLevel(logging.INFO)

    app_log = logging.getLogger("root")
    app_log.setLevel(logging.INFO)
    app_log.addHandler(my_handler)
    return app_log


def get_kafka():
    if get_env_bool("HIQ_OCI_STREAMING"):
        from hiq.vendor_oci_streaming import OciStreamingClient

        return OciStreamingClient()
    return None


class Jack(object):
    """Jack is a lumberjack to send trees to remote HiQ server in his own process space
    Jack is disabled by default. To enable it, set env variable JACK=1:
        export JACK=1
    """

    @staticmethod
    def consumer_func(queue, lock):
        if os.cpu_count() >= 2 and os.uname().sysname=='Linux':
            affinity_list = list(os.sched_getaffinity(0))
            os.sched_setaffinity(0, set(affinity_list[len(affinity_list) // 2 :]))

        pid = os.getpid()
        logger = log_jack_rotated()
        kafka_client = get_kafka()
        with lock:
            print("🅹 🅰 🅒 Ⓚ {} is started".format(pid))
        while True:
            try:
                data = queue.get()
                for key, value in data.items():
                    """
                    tree = Tree(value)
                    itree._itree.consolidate(tree.root)
                    # tree.show()
                    data = tree.repr()
                    """
                    if logger:
                        logger.info(key + "," + value)
                    if kafka_client:
                        kafka_client.produce_messages(key, value)
                    # with lock:
                    #    print(k, t)
                    #    print("🅹 {} got {} {}".format(pid, key, data))
            except Exception as e:
                time.sleep(0.1)
                print(e)

    def __init__(sf, *args, **kwargs):
        sf.invite_jack = get_env_bool("JACK")
        if not sf.invite_jack:
            sf.queue_jack = sf.consumer = None
            return
        sf.queue_jack = Queue()
        sf.lock = Lock()
        sf.consumer = Process(target=Jack.consumer_func, args=(sf.queue_jack, sf.lock))
        # This is critical! The consumer function has an infinite loop
        # Which means it will never exit unless we set daemon to true
        sf.consumer.daemon = True
        sf.consumer.start()

    def __del__(sf):
        if sf.consumer:
            sf.consumer.join()

    @_check_overhead
    def send_trees_to_jack(sf, d: dict, debug=False):
        if not sf.queue_jack:
            if debug:
                print("Jack is working")
            return
        data = {}
        for k, val in d.items():
            data[k] = val.repr()
        sf.queue_jack.put_nowait(data)
